#!/usr/bin/env python
# coding: utf-8
"""
This module tests if uploading data in chunks works as
expected (we should get the correct DataFrame length back).
"""
import pandas as pd
import pytest
from sqlalchemy import VARCHAR
from pangres import upsert
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import drop_table_for_test, read_example_table_from_db, TableNames


# # Helpers

@drop_table_for_test(TableNames.VARIOUS_CHUNKSIZES)
def insert_chunks(engine, schema, chunksize, nb_rows):
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)
    # MySQL does not want flexible text length in indices/PK
    dtype = {'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    upsert(schema=schema, table_name=TableNames.VARIOUS_CHUNKSIZES, df=df, chunksize=chunksize,
           con=engine, if_row_exists='update', dtype=dtype)
    df_db = read_example_table_from_db(engine=engine, schema=schema, table_name=TableNames.VARIOUS_CHUNKSIZES)
    # sort index (for MySQL...)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


# # Tests

@pytest.mark.parametrize('chunksize, nb_rows', [[1, 11], [3, 11]], ids=['one_by_one', 'odd_chunksize'])
def test_insert_various_chunksizes(engine, schema, chunksize, nb_rows):
    insert_chunks(engine, schema, chunksize=chunksize, nb_rows=nb_rows)



#!/usr/bin/env python
# coding: utf-8
import datetime
import logging
import numpy as np
import pandas as pd
import pytest
from sqlalchemy import create_engine, text, VARCHAR
from sqlalchemy.sql.elements import Null as SqlaNull
from pangres.examples import _TestsExampleTable
from pangres.exceptions import (DuplicateLabelsException,
                                DuplicateValuesInIndexException,
                                HasNoSchemaSystemException,
                                MissingIndexLevelInSqlException,
                                UnnamedIndexLevelsException)
from pangres.engine import PandasSpecialEngine
from pangres.helpers import _sqla_gt14
from pangres.tests.conftest import (commit, drop_table_for_test,
                                    get_table_namespace,
                                    schema_for_testing_creation,
                                    TableNames)


# # Test methods and attributes

# +
@drop_table_for_test(TableNames.NO_TABLE)
def test_repr(engine, schema):
    dummy_df = pd.DataFrame(index=pd.Index(data=[0], name='id'))
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, schema=schema,
                                  table_name=TableNames.NO_TABLE, df=dummy_df)
        # make sure it is printable without errors
        txt = str(pse)
        print(txt)
        # test some strings we expect to find in the repr
        for s in ('PandasSpecialEngine', 'id ', 'hexid', 'connection',
                  'schema', 'table', 'SQLalchemy table model'):
            assert s in txt


@drop_table_for_test(TableNames.NO_TABLE)
def test_table_attr(engine, schema):
    # generate a somewhat complex table model via the _TestsExampleTable class
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    table_name = TableNames.NO_TABLE
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, schema=schema,
                                  table_name=table_name, df=df)
        # make sure columns and table name match
        expected_cols = list(df.index.names) + df.columns.tolist()
        assert all((col in pse.table.columns for col in expected_cols))
        assert pse.table.name == table_name


@drop_table_for_test(TableNames.NO_TABLE)
def test_schema_creation(engine, schema):
    # overwrite schema
    schema = schema_for_testing_creation

    # local helper
    def drop_schema():
        if 'postgres' in engine.dialect.dialect_description:
            with engine.connect() as connection:
                connection.execute(text(f'DROP SCHEMA IF EXISTS {schema};'))
                commit(connection)

    # schema may already exist before testing
    drop_schema()

    # then try to create a schema from a PandasSpecialEngine instance
    dummy_df = pd.DataFrame(index=pd.Index(data=[0], name='id'))
    with engine.connect() as connection:
        try:
            pse = PandasSpecialEngine(connection=connection, schema=schema,
                                      table_name=TableNames.NO_TABLE, df=dummy_df)
            # this should raise HasNoSchemaSystemException
            # if we are not on a postgres engine
            assert not pse.schema_exists()
            pse.create_schema_if_not_exists()
            commit(connection)
            assert pse.schema_exists()
        except Exception as e:
            if pse._db_type == 'postgres' or not isinstance(e, HasNoSchemaSystemException):
                raise e
        finally:
            if pse._db_type == 'postgres':
                drop_schema()


@drop_table_for_test(TableNames.TABLE_CREATION)
def test_table_creation(engine, schema):
    dtype = {'profileid':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, schema=schema, dtype=dtype,
                                  table_name=TableNames.TABLE_CREATION, df=df)
        assert not pse.table_exists()
        pse.create_table_if_not_exists()
        commit(connection)
        assert pse.table_exists()


@pytest.mark.parametrize('axis', ['index', 'column'])
@drop_table_for_test(TableNames.ADD_NEW_COLUMN)
def test_add_new_columns(engine, schema, axis):
    # store arguments we will use for multiple PandasSpecialEngine instances
    table_name = TableNames.ADD_NEW_COLUMN
    common_kwargs = dict(schema=schema, table_name=table_name)
    common_kwargs['dtype'] = {'profileid':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # create our example table
    df = _TestsExampleTable.create_example_df(nb_rows=10)
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        pse.create_table_if_not_exists()
        commit(connection)
        assert pse.table_exists()

    # we need to recreate an instance of PandasSpecialEngine
    # so that a new table model with the new columns is created then add columns
    with engine.connect() as connection:
        # error message if we get unexpected values for "axis"
        # or we make a typo in our if/elif statements
        err_msg = f'Expected axis to be one of index, columns. Got {axis}'
        # add a new index level or new columns (no JSON ones,
        # it's not supported by sqlalchemy compilers :( )
        if axis == 'index':
            df['new_index_col'] = 'foo'
            df.set_index('new_index_col', append=True, inplace=True)
        elif axis == 'column':
            df = df.assign(new_text_col='test',
                           new_int_col=0,
                           new_float_col=1.1,
                           new_bool_col=False,
                           new_dt_col=pd.Timestamp('2020-01-01'),
                           # create this col for later
                           empty_col=None)
        else:
            raise AssertionError(err_msg)

        # recreate PandasSpecialEngine
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)

        # check if we get an error when trying to add an index level
        if axis == 'index':
            with pytest.raises(MissingIndexLevelInSqlException) as exc_info:
                pse.add_new_columns()
            assert 'Cannot add' in str(exc_info.value)
            return
        elif axis == 'column':
            pse.add_new_columns()
            commit(connection)
        else:
            raise AssertionError(err_msg)

    # check if the columns were correctly added
    # since we issued a return for 'index' earlier
    # the axis must now be 'columns'
    assert axis == 'column'
    # check the columns where added
    with engine.connect() as connection:
        ns = get_table_namespace(schema=schema, table_name=table_name)
        df_db = pd.read_sql(text(f'SELECT * FROM {ns} LIMIT 0;'), con=connection,
                            index_col='profileid')
        assert set(df.columns) == set(df_db.columns)


@pytest.mark.parametrize("new_empty_column_value", [1, 1.1, pd.Timestamp("2020-01-01", tz='UTC'),
                                                    {'foo':'bar'}, ['foo'], True])
@drop_table_for_test(TableNames.CHANGE_EMPTY_COL_TYPE)
def test_change_column_type_if_column_empty(engine, schema, caplog, new_empty_column_value):
    # store arguments we will use for multiple PandasSpecialEngine instances
    table_name = TableNames.CHANGE_EMPTY_COL_TYPE
    common_kwargs = dict(schema=schema, table_name=table_name)
    common_kwargs['dtype'] = {'profileid':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # json like will not work for sqlalchemy < 1.4
    # also skip sqlite as it does not support such alteration
    json_like = isinstance(new_empty_column_value, (dict, list)) and not _sqla_gt14()
    if json_like or 'sqlite' in engine.dialect.dialect_description:
        pytest.skip()

    # create our example table
    df = pd.DataFrame({'profileid':['foo'], 'empty_col':[None]}).set_index('profileid')
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        pse.create_table_if_not_exists()
        commit(connection)
        assert pse.table_exists()

    # recreate an instance of PandasSpecialEngine with a new df (so the model gets refreshed)
    # the line below is a "hack" to set any type of element as a column value
    # without pandas trying to broadcast it. This is useful when passing a list or such
    df['empty_col'] = df.index.map(lambda x: new_empty_column_value)
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, **common_kwargs)
        with caplog.at_level(logging.INFO, logger='pangres'):
            pse.adapt_dtype_of_empty_db_columns()
        assert len(caplog.records) == 1
        assert 'Changed type of column empty_col' in caplog.text


def test_values_conversion(_):
    engine = create_engine('sqlite:///')
    row = {'id':0,
           'pd_interval':pd.Interval(left=0, right=5),
           'nan':np.nan,
           'nat':pd.NaT,
           'none':None,
           'pd_na':getattr(pd, 'NA', None),
           'ts':pd.Timestamp('2021-01-01')}
    df = pd.DataFrame([row]).set_index('id')
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, df=df, table_name=TableNames.NO_TABLE)
        values = pse._get_values_to_insert()
    converted_row = values[0]
    assert len(row) == len(converted_row)
    # iterate over the keys of the row and get the index (after conversion we just have a list)
    for ix, k in enumerate(row):
        v = row[k]
        v_converted = converted_row[ix]
        if k == 'id':
            assert v_converted == v
        elif k == 'pd_interval':
            assert isinstance(v_converted, str)
        elif k == 'ts':
            assert isinstance(v_converted, datetime.datetime)
        # we should receive all null likes here
        else:
            assert pd.isna(v)
            assert isinstance(v_converted, SqlaNull)


# dummy connection string to test our categorization for databases
params_db_type_tests = [('sqlite:///', 'sqlite'),
                        ('postgresql+psycopg2://username:password@localhost:5432/postgres', 'postgres'),
                        ('postgresql://username:password@localhost:5432/postgres', 'postgres'),
                        ('mysql+pymysql://username:password@localhost:3306/db', 'mysql'),
                        ('oracle+cx_oracle://username:password@localhost', 'other')]


@pytest.mark.parametrize("connection_string, expected", params_db_type_tests)
def test_detect_db_type(_, connection_string, expected):
    engine = create_engine(connection_string)
    assert PandasSpecialEngine._detect_db_type(connectable=engine) == expected


# -

# # Test errors

# +
@drop_table_for_test(TableNames.NO_TABLE)
def test_error_index_level_named(engine, schema):
    df = pd.DataFrame({'test':[0]})
    with pytest.raises(UnnamedIndexLevelsException) as excinfo:
        with engine.connect() as connection:
            PandasSpecialEngine(connection=connection, schema=schema, table_name=TableNames.NO_TABLE, df=df)
    assert "All index levels must be named" in str(excinfo.value)


@pytest.mark.parametrize("option", ['index and column collision', 'columns duplicated', 'index duplicated'])
@drop_table_for_test(TableNames.NO_TABLE)
def test_duplicated_names(engine, schema, option):
    df = pd.DataFrame({'test':[0]})
    if option == 'index and column collision':
        df.index.name = 'test'
    elif option == 'columns duplicated':
        df.index.name = 'ix'
        df = df[['test', 'test']]
    elif option == 'index duplicated':
        df = df.set_index(['test', 'test'])
    else:
        raise AssertionError(f'Unexpected value for param `option`: {option}')

    with pytest.raises(DuplicateLabelsException) as excinfo:
        with engine.connect() as connection:
            PandasSpecialEngine(connection=connection, schema=schema, table_name=TableNames.NO_TABLE, df=df)
    assert "Found duplicates across index and columns" in str(excinfo.value)


@drop_table_for_test(TableNames.NO_TABLE)
def test_non_unique_index(engine, schema):
    df = pd.DataFrame(index=pd.Index(data=[0, 0], name='ix'))
    with pytest.raises(DuplicateValuesInIndexException) as excinfo:
        with engine.connect() as connection:
            PandasSpecialEngine(connection=connection, schema=schema, table_name=TableNames.NO_TABLE, df=df)
    assert "The index must be unique" in str(excinfo.value)


@pytest.mark.parametrize("bad_chunksize_value", [0, -1, 1.2])
@drop_table_for_test(TableNames.NO_TABLE)
def test_bad_chunksize(engine, schema, bad_chunksize_value):
    df = pd.DataFrame({'test':[0]})
    df.index.name = 'id'
    with engine.connect() as connection:
        pse = PandasSpecialEngine(connection=connection, schema=schema, table_name=TableNames.NO_TABLE, df=df)
        with pytest.raises(ValueError) as excinfo:
            pse._create_chunks(values=[0], chunksize=bad_chunksize_value)
        assert "integer strictly above 0" in str(excinfo.value)



#!/usr/bin/env python
# coding: utf-8
"""
End to end test similar to the scenario proposed in the docstring
of `pangres.upsert`.

We will create a table and then insert with update and then ignore for the
`ON CONFLICT` clause.
"""
import pandas as pd
import pytest
from sqlalchemy import create_engine, text, VARCHAR
from sqlalchemy.exc import ProgrammingError, OperationalError
from pangres import upsert
from pangres.examples import _TestsExampleTable
from pangres.exceptions import HasNoSchemaSystemException
from pangres.tests.conftest import (commit,
                                    drop_table,
                                    drop_table_for_test,
                                    get_table_namespace,
                                    read_example_table_from_db,
                                    schema_for_testing_creation,
                                    TableNames)


# # Helpers

utc_ts = lambda s: pd.Timestamp(s, tz='UTC')

# # Config

table_name = 'test_upsert_end_to_end'


# # Test data

# +
# df from which we will create a SQL table
df = _TestsExampleTable.create_example_df(nb_rows=6)
# test nulls won't create any problems
# avoid nulls in boolean column though as this is unpractical with pandas in older versions
df.iloc[0,[ix for ix, col in enumerate(df.columns) if col != 'likes_pizza']] = None

# df from which we will upsert update or upsert ignore
# remove one record from above and add one
# so we know that 1) old records are still there, 2) new ones get added
# 3) we can check whether the update/ignore of existing records worked
df2 = _TestsExampleTable.create_example_df(nb_rows=7)
df2 = df2.iloc[1:]
# -

# # Expectations

# +
df_after_insert_update = pd.concat(objs=(df.loc[~df.index.isin(df2.index.tolist())], # everything that is not in df2
                                         df2))

df_after_insert_ignore = pd.concat(objs=(df,
                                         df2.loc[~df2.index.isin(df.index.tolist())])) # everything that is not in df


# -

# # Tests

# +
# after the table is created with a first `upsert`
# using `create_table=False` or `create_table=True` should both work
@pytest.mark.parametrize('create_table', [False, True], ids=['create_table_false', 'create_table_true'])
@pytest.mark.parametrize('if_row_exists, df_expected', [['update', df_after_insert_update],
                                                        ['ignore', df_after_insert_ignore]],
                         ids=['update', 'ignore'])
@drop_table_for_test(TableNames.END_TO_END)
def test_end_to_end(engine, schema, create_table, if_row_exists, df_expected):
    # config
    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    table_name = TableNames.END_TO_END
    # common kwargs for every time we use the upsert function
    common_kwargs = dict(if_row_exists=if_row_exists, dtype=dtype, table_name=table_name)
    read_table = lambda: read_example_table_from_db(engine=engine, schema=schema, table_name=table_name).sort_index()

    # 1. create table
    upsert(con=engine, schema=schema, df=df, create_table=True, **common_kwargs)
    pd.testing.assert_frame_equal(df, read_table())

    # 2. insert update/ignore
    upsert(con=engine, schema=schema, df=df2, create_table=create_table, **common_kwargs)
    pd.testing.assert_frame_equal(df_expected, read_table())


def test_bad_value_if_row_exists(_):
    df = pd.DataFrame({'id':[0]}).set_index('id')
    engine = create_engine('sqlite:///')
    with pytest.raises(ValueError) as excinfo:
        upsert(con=engine, df=df, table_name=TableNames.NO_TABLE, if_row_exists='test')
    assert 'must be "ignore" or "update"' in str(excinfo.value)


@drop_table_for_test(TableNames.ADD_NEW_COLUMN)
def test_add_new_column(engine, schema):
    # config
    table_name = TableNames.ADD_NEW_COLUMN
    ns = get_table_namespace(schema=schema, table_name=table_name)
    dtype = {'id':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = pd.DataFrame({'id':['foo']}).set_index('id')
    # common kwargs for all the times we use upsert
    common_kwargs = dict(con=engine, schema=schema, df=df, table_name=table_name,
                         if_row_exists='update', dtype=dtype)

    # 1. create table
    upsert(**common_kwargs)
    # 2. add a new column and repeat upsert
    df['new_column'] = 'bar'
    upsert(**common_kwargs, add_new_columns=True)
    # verify content matches
    with engine.connect() as connection:
        df_db = pd.read_sql(text(f'SELECT * FROM {ns}'), con=connection, index_col='id')
        pd.testing.assert_frame_equal(df, df_db)


@drop_table_for_test(TableNames.CHANGE_EMPTY_COL_TYPE)
def test_adapt_column_type(engine, schema):
    # skip for sqlite as it does not support such alteration
    if 'sqlite' in engine.dialect.dialect_description:
        pytest.skip()

    # config
    table_name = TableNames.CHANGE_EMPTY_COL_TYPE
    dtype = {'id':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None
    df = pd.DataFrame({'id':['foo'], 'empty_column':[None]}).set_index('id')
    # common kwargs for all the times we use upsert
    common_kwargs = dict(con=engine, schema=schema, df=df, table_name=table_name,
                         if_row_exists='update', dtype=dtype)

    # 1. create table
    upsert(**common_kwargs)
    # 2. add non string data in empty column and repeat upsert
    df['empty_column'] = 1
    upsert(**common_kwargs, adapt_dtype_of_empty_db_columns=True)


@drop_table_for_test(TableNames.NO_TABLE)
def test_cannot_insert_missing_table_no_create(engine, schema):
    """
    Check if an error is raised when trying to insert in a missing table
    and `create_table` is False.
    """
    df = pd.DataFrame({'id':[0]}).set_index('id')
    with pytest.raises((OperationalError, ProgrammingError)) as excinfo:
        upsert(con=engine, schema=schema, df=df, table_name=TableNames.NO_TABLE,
               if_row_exists='update', create_table=False)
    assert any(s in str(excinfo.value) for s in ('no such table', 'does not exist', "doesn't exist"))


@drop_table_for_test(TableNames.CREATE_SCHEMA_NONE)
def test_create_schema_none(engine, schema):
    """
    If `create_schema` is True in `pangres.upsert` but the schema is `None`
    we should not raise an error even if it is a database that does not
    support schemas
    """
    df = pd.DataFrame({'id':[0]}).set_index('id')
    upsert(con=engine, schema=None, df=df, if_row_exists='update', create_schema=True,
            table_name=TableNames.CREATE_SCHEMA_NONE, create_table=True)


def test_create_schema_not_none(engine, schema):
    # local helpers
    is_postgres = 'postgres' in engine.dialect.dialect_description

    def drop_schema():
        if not is_postgres:
            return
        with engine.connect() as connection:
            connection.execute(text(f'DROP SCHEMA IF EXISTS {schema};'))
            commit(connection)

    # overwrite schema
    schema = schema_for_testing_creation

    # config
    df = pd.DataFrame({'id':[0]}).set_index('id')
    table_name = TableNames.CREATE_SCHEMA_NOT_NONE

    # drop table before test (could not get my decorator to work with another schema
    # when having an optional arg schema=None due to variable scopes problems)
    if is_postgres:
        drop_table(engine=engine, schema=schema, table_name=table_name)
        drop_schema()

    try:
        upsert(con=engine, schema=schema, df=df, if_row_exists='update', create_schema=True,
               table_name=table_name, create_table=True)
        if not is_postgres:
            raise AssertionError('Expected the upsert to fail when not using a postgres database')
    except Exception as e:
        # for postgres this should have worked
        if is_postgres:
            raise e
        else:
            assert isinstance(e, HasNoSchemaSystemException)
    finally:
        # drop table and schema after test
        if is_postgres:
            drop_table(engine=engine, schema=schema, table_name=table_name)
            drop_schema()



#!/usr/bin/env python
# coding: utf-8
# +
"""
Configuration and helpers for the tests of pangres with pytest.
"""
import json
import pandas as pd
import sqlalchemy as sa
from functools import wraps
from inspect import signature
from sqlalchemy import create_engine, text
from typing import Union

from pangres.helpers import _sqla_gt14


# -

# # Helpers for other test modules

# +
def _get_function_param_value(sig, param_name, args, kwargs):
    """
    Helper to retrieve the value of a specific parameter
    from *args and **kwargs like when we wrap a function
    """
    param_ix = list(sig.parameters).index(param_name)
    if param_ix <= len(args) - 1:
        return args[param_ix]
    else:
        return kwargs[param_name]


def table_exists(connection, schema, table_name) -> bool:
    insp = sa.inspect(connection)
    if _sqla_gt14():
        return insp.has_table(schema=schema, table_name=table_name)
    else:
        return table_name in insp.get_table_names(schema=schema)


def select_table(engine, schema, table_name,
                 error_if_missing=True,
                 **read_sql_kwargs) -> Union[pd.DataFrame, None]:
    """
    Does a simple SELECT * FROM {table} and returns a DataFrame from that.
    Has an option to return None if the table does not exist and
    `error_if_missing` is False.
    """
    ns = f'{schema}.{table_name}' if schema is not None else table_name
    with engine.connect() as con:
        # check if the table is present
        if table_exists(connection=con, schema=schema, table_name=table_name):
            return pd.read_sql(text(f'SELECT * FROM {ns}'), con=con, **read_sql_kwargs)
        elif error_if_missing:
            raise AssertionError(f'Table {ns} does not exist')
        else:
            return None


def drop_table(engine, schema, table_name):
    namespace = f'{schema}.{table_name}' if schema is not None else table_name
    with engine.connect() as connection:
        # instead of using `DROP TABLE IF EXISTS` which produces warnings
        # in MySQL when the table does exist, we will just check if it does exist
        # and using a `DROP TABLE` query accordingly
        if not table_exists(connection=connection, schema=schema,
                            table_name=table_name):
            return
        # make sure to commit this!
        connection.execute(text(f'DROP TABLE {namespace};'))
        if hasattr(connection, 'commit'):
            connection.commit()


# thanks to https://stackoverflow.com/a/42581103
def drop_table_for_test(table_name, drop_before_test=True, drop_after_test=True):
    """
    Decorator for wrapping our tests functions. Expects the tests function
    to have the arguments "engine" and "schema" and will retrieve their values
    for dropping the table named `table_name`.

    Parameters
    ----------
    drop_before_test
        Whether to drop the table before executing the test
    drop_after_test
        Whether to drop the table after executing the test
    """
    assert drop_before_test or drop_after_test, 'One of drop_before_test or drop_after_test must be True'

    def sub_decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            # get engine and schema
            sig = signature(function)
            assert 'engine' in sig.parameters
            assert 'schema' in sig.parameters
            engine = _get_function_param_value(sig=sig, param_name='engine', args=args, kwargs=kwargs)
            schema = _get_function_param_value(sig=sig, param_name='schema', args=args, kwargs=kwargs)

            # before test
            if drop_before_test:
                drop_table(engine=engine, schema=schema, table_name=table_name)

            # test
            function(*args, **kwargs)

            # after test
            if drop_after_test:
                drop_table(engine=engine, schema=schema, table_name=table_name)
            return
        return wrapper
    return sub_decorator


def get_table_namespace(schema, table_name):
    return f'{schema}.{table_name}' if schema is not None else table_name


def commit(connection):
    if hasattr(connection, 'commit'):
        connection.commit()


class TableNames:
    ADD_NEW_COLUMN = 'test_add_new_column'
    BAD_COLUMN_NAMES = 'test_bad_column_names'
    BAD_TEXT = 'test_bad_text'
    CHANGE_EMPTY_COL_TYPE = 'test_change_empty_col_type'
    COLUMN_NAMED_VALUES = 'test_column_named_values'
    COMMIT_AS_YOU_GO = 'test_commit_as_you_go'
    COMMIT_OR_ROLLBACK_TRANS = 'test_commit_or_rollback_trans'
    CREATE_SCHEMA_NONE = 'test_create_schema_none'
    CREATE_SCHEMA_NOT_NONE = 'test_create_schema_not_none'
    END_TO_END = 'test_end_to_end'
    INDEX_ONLY_INSERT = 'test_index_only_insert'
    INDEX_WITH_NULL = 'test_index_with_null'
    MULTIINDEX = 'test_multiindex'
    REUSE_CONNECTION = 'test_reuse_connection'
    TABLE_CREATION = 'test_table_creation'
    UNIQUE_KEY = 'test_unique_key'
    VARIOUS_CHUNKSIZES = 'test_chunksize'
    WITH_YIELD = 'test_with_yield'
    WITH_YIELD_EMPTY = 'test_with_yield_empty'
    # case when we need a table name but we are not going to use it for
    # creating/updating a table (if you see it in the DB, something went wrong)
    NO_TABLE = 'test_no_table'


# name of the PostgreSQL schema used for testing schema creation
schema_for_testing_creation = 'pangres_create_schema_test'


# -

# ## Class TestDB

# +
def pytest_addoption(parser):
    parser.addoption('--sqlite_conn', action="store", type=str, default=None)
    parser.addoption('--pg_conn', action="store", type=str, default=None)
    parser.addoption('--mysql_conn', action="store", type=str, default=None)
    parser.addoption('--pg_schema', action='store', type=str, default=None)


def pytest_generate_tests(metafunc):
    # this is called for every test
    # if we see the parameters "engine" and "schema" in a function
    # then we will repeat the test for each engine
    func_params = signature(metafunc.function).parameters
    if not ('engine' in func_params and 'schema' in func_params):
        # I could not find any other way than to add a dummy
        # for executing a test only once (parameterize needs arguments)
        metafunc.parametrize('_', [''], scope='module')
        return

    # tests that we need to repeat for each engine + options (e.g. future)
    conn_strings = {'sqlite':metafunc.config.option.sqlite_conn,
                    'pg':metafunc.config.option.pg_conn,
                    'mysql':metafunc.config.option.mysql_conn}
    engines, schemas, ids = [], [], []
    for db_type, conn_string in conn_strings.items():
        if conn_string is None:
            continue
        schema = metafunc.config.option.pg_schema if db_type == 'pg' else None
        engine = create_engine(conn_string)
        schemas.append(schema)
        engines.append(engine)
        schema_id = '' if schema is None else f'_schema:{schema}'
        ids.append(f'{engine.url.drivername}{schema_id}')
        # for sqlalchemy 1.4+ use future=True to try the future sqlalchemy 2.0
        if _sqla_gt14():
            future_engine = create_engine(conn_string, future=True)
            schemas.append(schema)
            engines.append(future_engine)
            ids.append(f'{engine.url.drivername}{schema_id}_future')
    assert len(engines) == len(schemas) == len(ids)
    if len(engines) == 0:
        raise ValueError('You must provide at least one connection string (e.g. argument --sqlite_conn)!')
    metafunc.parametrize("engine, schema", list(zip(engines, schemas)), ids=ids, scope='module')


# -

# ## Function to read back from database data we inserted
# We need to apply a few modification for comparing DataFrames we get back from the DB and DataFrames we expect e.g. for JSON (with SQlite pandas reads it as string).

def read_example_table_from_db(engine, schema, table_name):
    def load_json_if_needed(obj):
        """
        For SQlite we receive strings back (or None) for a JSON column.
        For Postgres we receive lists or dicts (or None) back.
        """
        if isinstance(obj, str):
            return json.loads(obj)
        return obj
    namespace = f'{schema}.{table_name}' if schema is not None else table_name
    with engine.connect() as connection:
        df_db = (pd.read_sql(text(f'SELECT * FROM {namespace}'), con=connection, index_col='profileid')
                 .astype({'likes_pizza':bool})
                 .assign(timestamp=lambda df: pd.to_datetime(df['timestamp'], utc=True))
                 .assign(favorite_colors= lambda df: df['favorite_colors'].map(load_json_if_needed)))
    return df_db



#!/usr/bin/env python
# coding: utf-8
# +
import pytest
from math import floor
from sqlalchemy import VARCHAR

from pangres import upsert
from pangres.helpers import _sqlite_gt3_32_0
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import drop_table_for_test, drop_table
# -


# # Config

table_name = 'test_speed'


# # Helpers

# +
def create_or_upsert_with_pangres(engine, schema, if_row_exists, df, chunksize, **kwargs):
    # MySQL does not want flexible text length in indices/PK
    dtype={'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    upsert(con=engine, df=df, schema=schema, chunksize=chunksize,
           table_name=table_name, if_row_exists=if_row_exists,
           dtype=dtype, **kwargs)


def create_with_pandas(engine, schema, df):
    dtype={'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None

    # we need this for SQlite when using pandas table creation
    # since we cannot use more than X parameters in a parameterized query
    if 'sqlite' in engine.dialect.dialect_description:
        max_params = 32766 if _sqlite_gt3_32_0() else 999
        col_len = len(df.columns) + len(df.index.names)
        chunksize = floor(max_params / col_len)
    else:
        chunksize = None

    # create table
    df.to_sql(con=engine, schema=schema, name=table_name, method='multi',
              chunksize=chunksize, dtype=dtype)


# -

# # Tests

pytest_params = dict(argnames='nb_rows, rounds, iterations', argvalues=[[10, 5, 1], [1_000, 1, 1]],
                     ids=['many_little_inserts', 'big_insert'])

# ## "Normal" insert speed
#
# Case where the table does not exist yet so there is no pk to compare.


@pytest.mark.parametrize('library', ['pandas', 'pangres'])
@pytest.mark.parametrize(**pytest_params)
@drop_table_for_test(table_name=table_name)
def test_create_and_insert_speed(engine, schema, benchmark, library, nb_rows, rounds, iterations):
    # get a df
    # we don't test JSON as this is problematic with pandas
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows).drop(columns=['favorite_colors'])

    # prepare funcs for benchmark and then do the benchmark
    switch = {'pangres':lambda: create_or_upsert_with_pangres(engine=engine, schema=schema, if_row_exists='update',
                                                              df=df, chunksize=nb_rows),
              'pandas':lambda: create_with_pandas(engine=engine, schema=schema, df=df)}

    benchmark.pedantic(switch[library], setup=lambda: drop_table(engine=engine, schema=schema, table_name=table_name),
                       rounds=rounds, iterations=iterations)


# ## Upsert overwrite speed

# this feature is not available in pandas yet
@pytest.mark.parametrize('library', ['pangres'])
@pytest.mark.parametrize('if_row_exists', ['update', 'ignore'])
@pytest.mark.parametrize(**pytest_params)
@drop_table_for_test(table_name=table_name)
def test_upsert_speed(engine, schema, benchmark, library, nb_rows, rounds, iterations, if_row_exists):
    assert library == 'pangres'  # in case pandas changes and we forget to update the tests

    # get a df
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows).drop(columns=['favorite_colors'])

    # setup for test (create table with no rows)
    def setup():
        create_or_upsert_with_pangres(engine=engine, schema=schema, if_row_exists=if_row_exists,
                                      df=df.head(0), chunksize=nb_rows)

    # test func
    # insert update/ignore with `create_table=False` to maximise speed
    func = lambda: create_or_upsert_with_pangres(engine=engine, schema=schema, if_row_exists=if_row_exists,
                                                 df=df, chunksize=nb_rows, create_table=False)

    benchmark.pedantic(func, setup=setup, rounds=rounds, iterations=iterations)



# +
"""
This module tests different ways of using pangres by providing a connection
instead of an engine and by using transactions
"""
import pandas as pd
import pytest
from sqlalchemy import text, VARCHAR

# local imports
from pangres import upsert
from pangres.transaction import TransactionHandler
from pangres.tests.conftest import commit, drop_table_for_test, select_table, TableNames


# -

# # Tests

# +
@drop_table_for_test(TableNames.REUSE_CONNECTION)
def test_connection_usable_after_upsert(engine, schema):
    df = pd.DataFrame(index=pd.Index([0], name='ix'))
    with engine.connect() as con:
        # do some random upsert operation
        upsert(con=con, df=df, schema=schema,
               table_name=TableNames.REUSE_CONNECTION,
               if_row_exists='update')
        # attempt to reuse the connection
        result = con.execute(text('SELECT 1;')).scalar()
        assert result == 1
        commit(con)


@pytest.mark.parametrize("trans_op", ['commit', 'rollback'])
@drop_table_for_test(TableNames.COMMIT_OR_ROLLBACK_TRANS)
def test_transaction(engine, schema, trans_op):
    df = pd.DataFrame(index=pd.Index(['foo'], name='ix'))
    table_name = TableNames.COMMIT_OR_ROLLBACK_TRANS
    # common keyword arguments for multiple upsert operations below
    common_kwargs = dict(schema=schema, table_name=table_name,
                         if_row_exists='update', dtype={'ix':VARCHAR(3)})

    with engine.connect() as con:
        trans = con.begin()
        try:
            # do some random upsert operation
            upsert(con=con, df=df, **common_kwargs)
            # do some other operation that requires commit
            upsert(con=con, df=df.rename(index={'foo':'bar'}), **common_kwargs)
            getattr(trans, trans_op)()  # commit or rollback
        finally:
            trans.close()

    # if trans_op=='commit': make sure we have "bar" and "foo" in the index
    # elif trans_op=='rollback': make sure we don't have any data
    # or that the table was not even created (what is rolled back
    # depends on the database type and other factors)
    if trans_op == 'commit':
        df_db = select_table(engine=engine, schema=schema, table_name=table_name, index_col='ix')
        pd.testing.assert_frame_equal(df_db.sort_index(),
                                      pd.DataFrame(index=pd.Index(['bar', 'foo'], name='ix')))
    elif trans_op == 'rollback':
        df_db = select_table(engine=engine, schema=schema, table_name=table_name, error_if_missing=False)
        # no table or an empty table
        assert df_db is None or len(df_db) == 0


@drop_table_for_test(TableNames.COMMIT_AS_YOU_GO)
def test_commit_as_you_go(engine, schema):
    df = pd.DataFrame(index=pd.Index(['foo'], name='ix'))
    table_name = TableNames.COMMIT_AS_YOU_GO
    # common keyword arguments for multiple upsert operations below
    common_kwargs = dict(schema=schema, table_name=table_name,
                         if_row_exists='update', dtype={'ix':VARCHAR(3)})

    with engine.connect() as con:
        # skip for sqlalchemy < 2.0 or when future=True flag is not passed
        # during engine creation (commit-as-you-go is a new feature)
        # when this is the case there is no attribute commit or rollback for
        # the connection
        if not hasattr(con, 'commit'):
            pytest.skip()

        # do some random upsert operation and commit
        upsert(con=con, df=df, **common_kwargs)
        con.commit()

        # do some other operation that requires commit and then rollback
        upsert(con=con, df=df.rename(index={'foo':'bar'}), **common_kwargs)
        con.rollback()

    # the table in the db should be equal to the initial df as the second
    # operation was rolled back
    df_db = select_table(engine=engine, schema=schema, table_name=table_name, index_col='ix')
    pd.testing.assert_frame_equal(df_db, df)


# -

# # Test errors

def test_non_connectable_transaction_handler(_):
    with pytest.raises(TypeError) as exc_info:
        with TransactionHandler(connectable='abc'):
            pass
    assert 'sqlalchemy connectable' in str(exc_info)



from pangres.core import aupsert, upsert
from pangres.utils import adjust_chunksize, fix_psycopg2_bad_cols
from pangres.examples import DocsExampleTable
from pangres._version import __version__
from pangres.exceptions import *


# # Versions checking

# +
def _py_gt3_10() -> bool:
    """
    Returns True if we are running Python >= 3.10 else return False.
    We need that due to different implementations for checking
    versions of other libraries
    """
    import sys
    py_v = sys.version_info
    # first condition is in case Python 4 comes out :P
    return (py_v.major > 3) or ((py_v.major >= 3) and (py_v.minor >= 10))


def _version_equal_or_greater_than(version_string, minimal_version_string) -> bool:
    if _py_gt3_10():
        from packaging import version
        return version.parse(version_string) >= version.parse(minimal_version_string)
    else:
        from distutils.version import LooseVersion
        return LooseVersion(version_string) >= LooseVersion(minimal_version_string)


def _sqla_gt14() -> bool:
    """
    Checks if sqlalchemy.__version__ is at least 1.4.0, when several
    deprecations were made.

    Stolen from pandas.io.sql (we don't import it as it's private
    and has just 2 lines of code).
    """
    import sqlalchemy
    return _version_equal_or_greater_than(version_string=sqlalchemy.__version__,
                                          minimal_version_string='1.4.0')


def _sqlite_gt3_32_0() -> bool:
    """
    Checks if the SQLite version is >= than 3.32.0.
    Starting from this version we can use more SQL parameters.
    See https://github.com/ThibTrip/pangres/issues/43
    """
    import sqlite3
    return _version_equal_or_greater_than(version_string=sqlite3.sqlite_version,
                                          minimal_version_string='3.32.0')


# -

# # Parameters checking

def validate_chunksize_param(chunksize:int):
    if not isinstance(chunksize, int):
        raise TypeError(f'Expected chunksize to be an int. Got {type(chunksize)}')
    if chunksize <= 0:
        raise ValueError('chunksize must be strictly above 0')



#!/usr/bin/env python
# coding: utf-8
"""
Here we test different indices:
1. MultiIndex
2. Indices with NULL

# Note
Other problems with indices such as duplicated values are handled by pangres.helpersPandasSpecialEngine.
"""
import pandas as pd
import pytest
from sqlalchemy import VARCHAR, text
from sqlalchemy.exc import IntegrityError
from pangres import upsert
from pangres.tests.conftest import drop_table_for_test, get_table_namespace, TableNames


# # Config

# +
# for creating table with MultiIndex
index_col = ['ix1', 'ix2', 'ix3']
df_multiindex = pd.DataFrame({'ix1': [1, 1], 'ix2': ['test', 'test2'],
                              'ix3': [pd.Timestamp('2019-01-01'), pd.Timestamp('2019-01-02')],
                              'foo': [1, 2]}).set_index(index_col)

# for inserting values
df_multiindex2 = pd.DataFrame({'ix1': [2, 2], 'ix2': ['test', 'test2'],
                               'ix3': [pd.Timestamp('2019-01-01'), pd.Timestamp('2019-01-02')],
                               'foo': [1, 2]}).set_index(index_col)


# -

# # Tests

# ## Test MultiIndex

@drop_table_for_test(TableNames.MULTIINDEX)
def test_create_and_insert_table_multiindex(engine, schema):
    # local helpers
    table_name = TableNames.MULTIINDEX
    namespace = get_table_namespace(schema=schema, table_name=table_name)

    def read_from_db():
        with engine.connect() as connection:
            df = pd.read_sql(text(f'SELECT * FROM {namespace}'), con=connection)
            df['ix3'] = pd.to_datetime(df['ix3'])
            return df.set_index(index_col)

    # dtype for index for MySQL... (can't have flexible text length)
    dtype = {'ix2':VARCHAR(5)} if 'mysql' in engine.dialect.dialect_description else None

    # create
    upsert(con=engine, schema=schema, df=df_multiindex, table_name=table_name, dtype=dtype, if_row_exists='update')
    df_db = read_from_db()
    pd.testing.assert_frame_equal(df_db, df_multiindex)

    # insert
    upsert(con=engine, schema=schema, df=df_multiindex2, table_name=table_name, dtype=dtype, if_row_exists='update')
    df_db = read_from_db()
    pd.testing.assert_frame_equal(df_db, pd.concat(objs=[df_multiindex, df_multiindex2]))


# ## Test index with null value

@drop_table_for_test(TableNames.INDEX_WITH_NULL)
def test_index_with_null(engine, schema):
    df = pd.DataFrame({'ix':[None, 0], 'test': [0, 1]}).set_index('ix')
    with pytest.raises(IntegrityError):
        upsert(con=engine, schema=schema, df=df, table_name=TableNames.INDEX_WITH_NULL, if_row_exists='update')
        # don't test error for mysql since only a warning is raised and the line is skipped
        if 'mysql' in engine.dialect.dialect_description:
            pytest.skip()


# ## Test only index

@pytest.mark.parametrize("if_row_exists", ['update', 'ignore'])
@drop_table_for_test(TableNames.INDEX_ONLY_INSERT)
def test_only_index(engine, schema, if_row_exists):
    # config
    table_name = TableNames.INDEX_ONLY_INSERT

    # upsert df with only index
    df = pd.DataFrame({'ix':[1]}).set_index('ix')
    upsert(con=engine, schema=schema, df=df, table_name=table_name, if_row_exists=if_row_exists)

    # check data integrity
    namespace = get_table_namespace(schema=schema, table_name=table_name)
    with engine.connect() as connection:
        df_db = pd.read_sql(text(f'SELECT * FROM {namespace}'), con=connection)
    assert 'ix' in df_db.columns
    assert len(df_db) > 0
    assert df_db['ix'].iloc[0] == 1



#!/usr/bin/env python
# coding: utf-8
"""
This module tests we can get information back from
the upserted chunks when the parameter `yield_chunks`
is True. It also check the integrity of the data.
"""
import math
import pandas as pd
from sqlalchemy import VARCHAR, INT
from pangres import upsert
from pangres.examples import _TestsExampleTable
from pangres.tests.conftest import (drop_table_for_test,
                                    read_example_table_from_db,
                                    TableNames)


# # Insert values one by one

@drop_table_for_test(TableNames.WITH_YIELD)
def test_get_nb_rows(engine, schema):
    # config
    table_name = TableNames.WITH_YIELD
    nb_rows, chunksize = 20, 3
    nb_last_chunk = nb_rows % chunksize
    nb_chunks = math.ceil(nb_rows/chunksize)
    # MySQL does not want flexible text length in indices/PK
    dtype = {'profileid':VARCHAR(10)} if 'mysql' in engine.dialect.dialect_description else None
    df = _TestsExampleTable.create_example_df(nb_rows=nb_rows)

    # iterate over upsert results
    # make sure we can extract the number of updated rows and that it is correct
    iterator = upsert(con=engine, df=df, table_name=table_name, if_row_exists='update',
                      schema=schema, chunksize=chunksize, dtype=dtype, yield_chunks=True)

    for ix, result in enumerate(iterator):
        assert result.rowcount == (chunksize if ix != nb_chunks-1 else nb_last_chunk)

    # verify the inserted data is as expected
    # we sort the index for MySQL
    df_db = read_example_table_from_db(engine=engine, schema=schema, table_name=table_name)
    pd.testing.assert_frame_equal(df.sort_index(), df_db.sort_index())


# # Test of an empty DataFrame

@drop_table_for_test(TableNames.WITH_YIELD_EMPTY)
def test_yield_empty_df(engine, schema):
    df = pd.DataFrame({'id':[], 'value':[]}).set_index('id')

    # we should get an empty generator back
    iterator = upsert(con=engine, df=df, table_name=TableNames.WITH_YIELD_EMPTY, if_row_exists='update',
                      schema=schema, dtype={'id':INT, 'value':INT}, yield_chunks=True)

    # the for loop should never run because the generator should be empty
    for result in iterator:
        raise AssertionError('Expected the generator returned by upsert with an empty df to be empty')



# -*- coding: utf-8 -*-
# +
import pandas as pd
import pytest
import random

from pangres import upsert, fix_psycopg2_bad_cols
from pangres.exceptions import BadColumnNamesException
from pangres.tests.conftest import drop_table_for_test, TableNames
# -

# # Helpers

# list of characters that could potentially cause errors in column names or even values
bad_char_seq = """/_- ?§$&"',:;*()%[]{}|<>=!+#""" + "\\"


# # Add colums with bad names and insert bad text values

# +
@drop_table_for_test(TableNames.BAD_TEXT)
def test_bad_text_insert(engine, schema):
    # add bad text in a column named 'text'
    create_random_text = lambda: ''.join(random.choice(bad_char_seq) for i in range(10))
    df_test = (pd.DataFrame({'text': [create_random_text() for i in range(10)]})
               .rename_axis(['profileid'], axis='index', inplace=False))
    upsert(con=engine, schema=schema, table_name=TableNames.BAD_TEXT, df=df_test,
           if_row_exists='update')


# do the next test multiple times to try different combinations of bad characters
@pytest.mark.parametrize('iteration', range(5), ids=[f'iteration{i}' for i in range(5)])
@drop_table_for_test(TableNames.BAD_COLUMN_NAMES)
def test_bad_column_names(engine, schema, iteration):
    # add columns with bad names
    # don't do this for MySQL which has more strict rules for column names
    if 'mysql' in engine.dialect.dialect_description:
        pytest.skip()

    random_bad_col_name = ''.join(random.choice(bad_char_seq) for i in range(50))
    df_test = (pd.DataFrame({random_bad_col_name: ['test', None]})
               .rename_axis(['profileid'], axis='index', inplace=False))

    # psycopg2 can't process columns with "%" or "(" or ")" so we will need `fix_psycopg2_bad_cols`
    if 'postgres' in engine.dialect.dialect_description:
        df_test = fix_psycopg2_bad_cols(df_test)
    upsert(con=engine, schema=schema, df=df_test, table_name=TableNames.BAD_COLUMN_NAMES, if_row_exists='update')


@drop_table_for_test(TableNames.BAD_COLUMN_NAMES)
def test_bad_column_name_postgres_raises(engine, schema):
    if 'postgres' not in engine.dialect.dialect_description:
        pytest.skip()
    df = pd.DataFrame({'id':[0], '(test)':[0]}).set_index('id')
    with pytest.raises(BadColumnNamesException) as exc_info:
        upsert(con=engine, schema=schema, df=df, table_name=TableNames.BAD_COLUMN_NAMES, if_row_exists='update')
    assert 'does not seem to support column names with' in str(exc_info.value)


# -

# # Another test with the column name `values` (see issue #34 of pangres)

@drop_table_for_test(TableNames.COLUMN_NAMED_VALUES)
def test_column_named_values(engine, schema):
    df = pd.DataFrame({'values': range(5, 9)}, index=pd.Index(range(1, 5), name='idx'))
    upsert(con=engine, schema=schema, df=df, if_row_exists='update', table_name=TableNames.COLUMN_NAMED_VALUES)



# +
import random
import datetime
import pandas as pd
from sqlalchemy import (Column, TEXT, FLOAT, BOOLEAN,
                        JSON, VARCHAR, DATETIME)
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

class _TestsExampleTable(Base):
    """
    Example table compatible with Postgres, SQLite and MySQL for testing.
    """
    __tablename__ = 'pangres_example'
    # use VARCHAR for the pk so MySQL doesn't complain...
    # MySQL does not want variable length text as index
    profileid = Column(VARCHAR(10), primary_key=True)
    email = Column(TEXT)
    timestamp = Column(DATETIME(timezone=True))
    size_in_meters = Column(FLOAT)
    likes_pizza = Column(BOOLEAN)
    favorite_colors = Column(JSON)

    @classmethod
    def create_example_df(cls, nb_rows):
        emails = ['foo', 'bar', 'baz', 'test', 'abc', 'foobar', 'foobaz']
        domains = ['gmail.com', 'yahoo.fr', 'yahoo.com', 'outlook.fr']
        email_choices = []
        for i in range(nb_rows):
            email = random.choice(emails)
            domain = random.choice(domains)
            email_choices.append(f'{email}@{domain}')
        timestamps = [(datetime.datetime
                       .fromtimestamp(random.randint(1000000000,1300000000))
                       .astimezone(datetime.timezone.utc))
                      for i in range(nb_rows)]
        colors = ['yellow', 'blue', 'pink', 'red', 'orange', 'brown']
        favorite_colors = []
        for i in range(nb_rows):
            l = [random.choice(colors) for i in range(random.randint(1,3))]
            favorite_colors.append(l)
        data = {# IMPORTANT! for our tests make profileid incremental!
                # it's not an integer (see table definition for an explanation why not)
                # but we just add a prefix or something to an incremented number
                'profileid':[str(f'abc{i}') for i in range(nb_rows)],
                'email':email_choices,
                'timestamp':timestamps,
                'size_in_meters':[random.uniform(1.5,2.3) for i in range(nb_rows)],
                'likes_pizza':[random.choice([True, False]) for i in range(nb_rows)],
                'favorite_colors':favorite_colors}
        df = pd.DataFrame(data).set_index('profileid')
        return df

class DocsExampleTable():
    """
    Example DataFrames for the docs.
    """
    # create some test data
    _data = {'full_name':['John Rambo', 'The Rock', 'John Travolta'],
             'likes_sport':[True, True, False],
             'updated':[pd.Timestamp('2020-02-01', tz='UTC'),
                        pd.Timestamp('2020-04-01', tz='UTC'), pd.NaT],
             'size_in_meters':[1.77, 1.96, None]}
    # create DataFrame using this test data
    df = pd.DataFrame(_data).set_index('full_name')
    # create test data for showing an INSERT UPDATE
    _new_data = {'full_name':['John Travolta', 'Arnold Schwarzenegger'],
                 'likes_sport':[True, True],
                 'updated':[pd.Timestamp('2020-04-04', tz='UTC'), pd.NaT],
                 'size_in_meters':[1.88, 1.88]}
    new_df =  pd.DataFrame(_new_data).set_index('full_name')
    # create test data for showing an INSERT IGNORE
    _new_data2 = {'full_name':['John Travolta', 'John Cena'],
                  'likes_sport':[True, True],
                  'updated':[pd.NaT, pd.NaT],
                  'size_in_meters':[2.50, 1.84]}
    new_df2 =  pd.DataFrame(_new_data2).set_index('full_name')

    # test DataFrame for the methods of upsert.UpsertQuery
    df_upsert = pd.DataFrame(index=pd.Index(data=['foo', 'bar', 'baz'], name='ix'))
    df_upsert['email'] = ['abc@outlook.fr', 'baz@yahoo.fr', 'foobar@gmail.com']
    df_upsert['ts'] = [pd.Timestamp('2021-01-01', tz='UTC')]*3
    df_upsert['float'] = [1.1, 1.2, 1.3]
    df_upsert['bool'] = [True, False, False]
    df_upsert['json'] = [['red', 'yellow'], ['yellow'], ['yellow', 'red']]    
